stopifnot(require(tools, quietly = TRUE))

.is_package <- function(package, packages) is.element(el = package, set = packages[, "Package"])

.is_package_installed <- function(installed_packages) {
  function(package) .is_package(package, installed_packages)
}

.is_old_package <- function(old_packages) {
  function(package) .is_package(package, old_packages)
}

.is_package_dependencies_installed <- function(package_dependencies, installed_packages) {
  function(package) {
    dependencies <- package_dependencies[[package]]
    dependencies_installed <- sapply(dependencies, .is_package_installed(installed_packages))
    all(dependencies_installed)
  }
}

.is_bioconductor_package <- function(available_packages) {
  function(package) ifelse(!package %in% row.names(available_packages), FALSE, grepl("bioconductor.org", available_packages[package, "Repository"]))
}

.repositories <- function(repos) {
  # BiocManager::repositories() use the repos option, see https://rdrr.io/cran/BiocManager/src/R/repositories.R
  options(repos = repos)
  tryCatch(BiocManager::repositories(), error = function(error) {
    message(error$message)
    repos
  })
}

.available_packages <- function(installed_packages, repos, type) {
  available_packages <- available.packages(repos = repos, type = type)
  if (nrow(available_packages) == 0) {
    # using installed_packages if none available_packages (no connection to repositories)
    available_packages <- installed_packages
  }
  return(available_packages)
}

.install_packages <- function(packages, lib, repos, available_packages, type) {
  bioconductor <- sapply(packages, .is_bioconductor_package(available_packages))
  if (!all(bioconductor)) {
    install.packages(packages[!bioconductor], lib = lib, repos = repos, type = type)
  }
  if (any(bioconductor)) {
    BiocManager::install(packages[bioconductor], lib = lib, update = FALSE, type = type)
  }
}

.installed_packages_with_dependencies <- function(packages, installed_packages, package_dependencies) {
  installed <- sapply(packages, .is_package_installed(installed_packages))
  dependencies_installed <- sapply(packages, .is_package_dependencies_installed(package_dependencies, installed_packages))
  package_dependencies_installed <- lapply(package_dependencies, function(dependencies) sapply(dependencies, .is_package_installed(installed_packages)))
  return(list(installed = all(installed & dependencies_installed), installed_packages = installed, installed_packages_dependencies = dependencies_installed, installed_dependencies = package_dependencies_installed))
}

#' Check if packages are installable
#'
#' @param packages character vector, the names of packages to check if there are installable
#' @param lib character, giving the library directories where to check the installed packages
#' @param repos character, base URL of the CRAN repository to use
#' @param type character, indicating the type of package to download and install
#'
#' @return list, describing installable status of packages
#' \describe{
#'   \item{installable}{logical, any package is installable}
#'   \item{installed_packages}{named logical, is package is installed}
#'   \item{installed_packages_dependencies}{named logical, is package dependencies are installed}
#'   \item{installed_dependencies}{named list of named logical, dependencies of packages that are installed}
#' }
#'
installable_packages <- function(packages, lib, repos, type) {
  # for installed.packages using .libPaths() instead of lib.loc to find default libraries installed inside .Library
  .libPaths(lib)

  repositories <- .repositories(repos)

  # check installed packages
  # noCache = TRUE (when lib is on network drive, the cache may be inconsistent, timestamp difference between local and network drives)
  installed_packages <- installed.packages(noCache = TRUE)
  installed <- sapply(packages, .is_package_installed(installed_packages))

  # check updatable packages
  available_packages <- .available_packages(installed_packages, repos = repositories, type = type)
  old_packages <- old.packages(lib.loc = lib, repos = repositories, instPkgs = installed_packages, available = available_packages, type = type)
  old <- sapply(packages, .is_old_package(old_packages))

  # check packages dependencies
  # with installed_packages is sufficient (search the first uninstalled dependency in installed_packages hierarchy)
  package_dependencies <- package_dependencies(packages, db = installed_packages, recursive = TRUE)
  dependencies <- unique(Reduce(c, package_dependencies))
  dependencies_installed <- sapply(dependencies, .is_package_installed(installed_packages))
  packages_dependencies_installed <- sapply(packages, .is_package_dependencies_installed(package_dependencies, installed_packages))

  installable <- any(!installed) | any(old) | any(!dependencies_installed)
  package_dependencies_installed <- lapply(package_dependencies, function(dependencies) sapply(dependencies, .is_package_installed(installed_packages)))
  return(list(installable = installable, installed_packages = installed, updatable_packages = old, installed_packages_dependencies = packages_dependencies_installed, installed_dependencies = package_dependencies_installed))
}

#' Install packages
#'
#' @param packages character vector, the names of packages to install
#' @param lib character, giving the library directories where to install the packages
#' @param repos character, base URL of the CRAN repository to use
#' @param type character, indicating the type of package to download and install
#'
#' @return list, describing installation status of packages
#' \describe{
#'   \item{installed}{logical, all packages are installed and their dependencies}
#'   \item{installed_packages}{named logical, is package is installed}
#'   \item{installed_packages_dependencies}{named logical, is package dependencies are installed}
#'   \item{installed_dependencies}{named list of named logical, dependencies of packages that are installed}
#' }
#'
install_packages <- function(packages, lib, repos, type) {
  # for installed.packages using .libPaths() instead of lib.loc to find default libraries installed inside .Library
  .libPaths(lib)

  repositories <- .repositories(repos)

  # check installed packages
  # noCache = TRUE (when lib is on network drive, the cache may be inconsistent, timestamp difference between local and network drives)
  installed_packages <- installed.packages(noCache = TRUE)
  installed <- sapply(packages, .is_package_installed(installed_packages))
  if (any(!installed)) {
    message(paste("Packages [", paste0(packages[!installed], collapse = ", "), "] are not installed"))
  }

  # check updatable packages
  available_packages <- .available_packages(installed_packages, repos = repositories, type = type)
  old_packages <- old.packages(lib.loc = lib, repos = repositories, instPkgs = installed_packages, available = available_packages, type = type)
  old <- sapply(packages, .is_old_package(old_packages))
  if (any(old)) {
    message(paste("Packages [", paste0(packages[old], collapse = ", "), "] are updatable"))
  }

  # process installable packages
  installable <- !installed | old
  if (any(installable)) {
    installable_packages <- packages[installable]
    message(paste("Installing packages [", paste0(installable_packages, collapse = ", "), "]"))
    .install_packages(installable_packages, lib, repos, available_packages, type)
    installed_packages <- installed.packages(noCache = TRUE)
    installed <- sapply(packages, .is_package_installed(installed_packages))
    if (any(!installed)) {
      message(paste("Installation packages [", paste0(packages[!installed], collapse = ", "), "] failed"))
      # return here, no needs to continue with packages dependencies installation if main packages are not installed
      package_dependencies <- package_dependencies(packages, db = available_packages, recursive = TRUE)
      return(.installed_packages_with_dependencies(packages, installed_packages, package_dependencies))
    }
  }

  # process packages dependencies installation
  package_dependencies <- package_dependencies(packages, db = available_packages, recursive = TRUE)
  dependencies <- unique(Reduce(c, package_dependencies))
  dependencies_installed <- sapply(dependencies, .is_package_installed(installed_packages))
  if (!all(dependencies_installed)) {
    installable_dependencies <- dependencies[!dependencies_installed]
    message(paste("Installing dependencies [", paste0(installable_dependencies, collapse = ", "), "]"))
    .install_packages(installable_dependencies, lib, repos, available_packages, type)
    installed_packages <- installed.packages(noCache = TRUE)
    dependencies_installed <- sapply(dependencies, .is_package_installed(installed_packages))
    if (any(!dependencies_installed)) {
      message(paste("Installation dependencies [", paste0(dependencies[!dependencies_installed], collapse = ", "), "] failed"))
    }
  }
  return(.installed_packages_with_dependencies(packages, installed_packages, package_dependencies))
}
