
.normalize <- function(x, range) {
  (x - range["min"]) / (range["max"] - range["min"])
}

.denormalize <- function(x, range) {
  range["min"] + x * (range["max"] - range["min"])
}

setGeneric(name = ".get_learning", def = function(learning, data) {
  standardGeneric(".get_learning")
})

setMethod(f = ".get_learning", signature = c("data.frame", "data.frame"), definition = function(learning, data) {
  return(data)
})

setMethod(f = ".get_learning", signature = c("SpatialPointsDataFrame", "data.frame"), definition = function(learning, data) {
  return(SpatialPointsDataFrame(learning@coords, data, proj4string = rebuild_CRS(learning@proj4string), match.ID = FALSE))
})

setMethod(f = ".get_learning", signature = c("SpatialMultiPointsDataFrame", "data.frame"), definition = function(learning, data) {
  return(SpatialMultiPointsDataFrame(learning@coords, data, proj4string = rebuild_CRS(learning@proj4string), match.ID = FALSE))
})

setMethod(f = ".get_learning", signature = c("SpatialPolygonsDataFrame", "data.frame"), definition = function(learning, data) {
  Sr <- SpatialPolygons(learning@polygons, proj4string = rebuild_CRS(learning@proj4string))
  return(SpatialPolygonsDataFrame(Sr, data, match.ID = FALSE))
})

setGeneric(name = ".aggregate_variables", def = function(source, variables) {
  standardGeneric(".aggregate_variables")
})

setMethod(f = ".aggregate_variables", signature = c("data.frame", "list"), definition = function(source, variables) {
  return(.aggregate_node(source, variables))
})

setMethod(f = ".aggregate_variables", signature = c("SpatialPointsDataFrame", "list"), definition = function(source, variables) {
  return(.aggregate_node(source@data, variables))
})

setMethod(f = ".aggregate_variables", signature = c("SpatialMultiPointsDataFrame", "list"), definition = function(source, variables) {
  return(.aggregate_node(source@data, variables))
})

setMethod(f = ".aggregate_variables", signature = c("SpatialPolygonsDataFrame", "list"), definition = function(source, variables) {
  return(.aggregate_node(source@data, variables))
})

learning_process <- R6Class("learning_process",
  private = list(
    .source = NULL,
    .attribute = NULL,
    .variables = NULL,
    .normalization_range = NULL,
    # number of digits to rounding weights
    .weight_digits = NULL,
    .operator = NULL,
    .observed = NULL,
    .infered = NULL,
    .error = NULL,
    .learning = NULL,
    .r2 = NULL,

    get_inputs = function() {
      variable_names <- sapply(private$.variables, function(node) node$name)
      .aggregate_variables(private$.source, private$.variables)[variable_names]
    }
  ),

  active = list(
    attribute = function(attribute) {
      private$.attribute <- attribute
      private$.observed <- private$.source[[attribute]]
    },

    variables = function(variables) private$.variables <- variables,

    normalization_range = function(normalization_range) {
      if (missing(normalization_range)) {
        private$.normalization_range
      } else {
        private$.normalization_range <- normalization_range
      }
    },

    weight_digits = function(weight_digits) private$.weight_digits <- weight_digits,

    operator = function(operator) private$.operator <- operator,

    observed = function() private$.observed,

    infered = function() private$.infered,

    error = function() private$.error,

    learning = function() private$.learning,

    r2 = function() private$.r2
  ),

  public = list(
    initialize = function(source) {
      private$.source <- source
      private$.weight_digits <- 3
    },

    get_owa_weights = function() {
      inputs <- private$get_inputs()
      observed <- .normalize(private$.observed, private$.normalization_range)
      weights <- LearnOwaWeights(inputs, observed)
    },

    get_wam_weights = function() {
      inputs <- private$get_inputs()
      observed <- .normalize(private$.observed, private$.normalization_range)
      weights <- LearnWamWeights(inputs, observed)
    },

    perform_inference = function() {
      # build aggregate Node
      aggregate <- Node$new("inferred")
      for (variable in private$.variables) {
        aggregate$AddChildNode(variable)
      }
      aggregate$aggreg <- private$.operator
      # compute inference
      learning <- .aggregate_node(private$.source, aggregate)
      private$.infered <- .denormalize(learning[["inferred"]], private$.normalization_range)
      private$.error <- private$.infered - private$.observed
      # build the learning data frame
      data <- data.frame(private$.infered, private$.observed, private$.error)
      colnames(data) <- c("inferred", "observed", "error")
      private$.learning <- .get_learning(learning, data)
      # compute r2
      private$.r2 <- cor(private$.infered, private$.observed)^2
    }
  )
)
