#' Build Polygons from coordinates and hole informations.
#'
#' @param x_coords numeric vector, x coordinates of all Polygon to build.
#' @param y_coords numeric vector, y coordinates of all Polygon to build.
#' @param points_in_polygons integer vector, number of points in each Polygon.
#' @param holes logical vector, for each Polygon is hole.
#' @param hole_ofs integer vector, index of the shell Polygon for each hole Polygon (same format as rgeos createPolygonsComment).
#'
#' @return a closure function to build Polygons.
#'
#' @export
#' @keywords internal
#'
make_polygons <- function(x_coords, y_coords, points_in_polygons, holes, hole_ofs) {
  polygons_index <- 1 # the Polygons id
  polygon_index <- 1 # the Polygon index
  hole_of_index <- 1 # the hole_of index
  coord_index <- 1 # the coordinate index

  #' Main function to build Polygon.
  #'
  #' @param dummy unused parameter.
  #'
  #' @return a Polygon object.
  #'
  make_polygon <- function(dummy) {
    points_in_polygon <- points_in_polygons[polygon_index]
    coord_end <- coord_index + points_in_polygon - 1
    x <- x_coords[coord_index:coord_end]
    y <- y_coords[coord_index:coord_end]
    coords <- cbind(x, y)
    hole <- holes[polygon_index]
    polygon_index <<- polygon_index + 1
    coord_index <<- coord_end + 1
    Polygon(coords = coords, hole = hole)
  }

  #' Main function to build Polygons, with comment (same format as rgeos createPolygonsComment).
  #' @seealso <https://www.rdocumentation.org/packages/sp/topics/SpatialPolygons>
  #' @seealso <https://www.rdocumentation.org/packages/rgeos/topics/RGEOS%20Polygon%20Hole%20Comment%20Functions>
  #'
  #' @param polygon_size integer, the number of Polygon inside the Polygons to build.
  #'
  #' @return a Polygons object.
  #'
  function(polygon_size) {
    srl <- lapply(c(1:polygon_size), make_polygon)
    polygons <- Polygons(srl = srl, ID = polygons_index)
    polygons_index <<- polygons_index + 1
    hole_of_end <- hole_of_index + polygon_size - 1
    comment <- paste(hole_ofs[hole_of_index:hole_of_end], collapse = " ")
    hole_of_index <<- hole_of_end + 1
    comment(polygons) <- comment
    polygons
  }
}

#' Explode the Polygon list of a Polygons object to a list of list of Polygon, each shell Polygon associated with his holes (based on Polygons comment).
#' @seealso <https://www.rdocumentation.org/packages/sp/topics/SpatialPolygons>
#' @seealso <https://www.rdocumentation.org/packages/rgeos/topics/RGEOS%20Polygon%20Hole%20Comment%20Functions>
#'
#' @param polygons Polygons object, the Polygons to explode.
#'
#' @return a list of list of Polygon (the first Polygon in the list is the shell Polygon, the other are holes).
#'
#' @keywords internal
#'
explode_polygons <- function(polygons) {
  comment <- comment(polygons)
  if (is.null(comment)) {
    sf_polygons <- st_as_sf(SpatialPolygons(list(polygons)))
    polygons <- as_Spatial(sf_polygons)@polygons[[1]]
    comment <- comment(polygons)
  }
  hole_ofs <- as.integer(unlist(strsplit(comment, " ")))
  hole_of_ids <- cbind(id = 1:length(hole_ofs), hole_of = hole_ofs)
  shell_ids <- hole_of_ids[hole_of_ids[, 2] == 0, 1]
  hole_of_ids <- hole_of_ids[hole_of_ids[, 2] > 0, , drop = FALSE] # drop = FALSE, keep always as matrix

  #' Explode the shell Polygon corresponding to the shell_id
  #'
  #' @param shell_id integer, the id of shell Polygon to explode.
  #'
  #' @return a list of Polygon (the first Polygon is the shell, the other are holes).
  #'
  explode_shell_polygons <- function(shell_id) {
    hole_ids <- hole_of_ids[hole_of_ids[, 2] == shell_id, 1] # find holes associated with the shell
    polygons@Polygons[c(shell_id, hole_ids)]
  }

  lapply(shell_ids, explode_shell_polygons)
}

#' Explode the Polygon list of a Polygons object to a list of list of numeric matrix, the coords of each Polygon, each shell Polygon associated with his holes.
#' @seealso [explode_polygons()]
#'
#' @param polygons Polygons object, the Polygons to explode.
#'
#' @return a list of list of numeric matrix (the first matrix in the list is coords of the shell Polygon, the other are holes coords).
#'
#' @export
#' @keywords internal
#'
explode_polygons_coords <- function(polygons) {
  #' Transform a list of Polygon to their coords.
  #'
  #' @param polygon_list list of Polygon.
  #'
  #' @return a list of numeric matrix.
  #'
  to_polygons_coords <- function(polygon_list) {
    lapply(polygon_list, function(polygon) polygon@coords)
  }

  exploded_polygons <- explode_polygons(polygons)
  lapply(exploded_polygons, to_polygons_coords)
}
